// Author: A.Voss@FH-Aachen.de
//
// initializing data and class-members used in C++11 !
// tags: brace-initializers

#include <iostream>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

// note that brace-initializers also work for local variables
// and not only in a class context, but here they are most useful
class C {
public:
    // define some default initializers
    int n = 23;             // default value if not overriden
    int m {42};             // another syntax
    int y,z{};              // y is not initialized, z to zero (0) 
    int a[3] = { 1,2,3 };   // instead of C() : a{1,2,3}
    int b[2] { 11,12 };     // with or without '='
    int *p = new int[2] { -2,-3 };  // even dynamic arrays
    
    // if you explicitly provide initializers, these are taken
    C() : b{3,4} { }
    ~C() { delete [] p; }
};

// how to provide a similar initializer
class D {
public:
    vector<double> v;
    
    // use initializer_list template with correct data type T
    D(const initializer_list<double>& l) // : v(l)
    {
        // example: manual access
        for (double x: l)
            v.push_back(x);
        // note: standard containers support init.lists by themself
        // use ": v(l)" in this case
    }
    
};

int main()
{
    C c;
    tools_log(); cout << "c.n=" << c.n << ", c.m=" << c.m << ", c.y=" << c.y << ", c.z=" << c.z << endl;
    tools_log(); cout << "c.a[2]=" << c.a[2] << ", c.b[1]=" << c.b[1] << endl;
    
    vector<int> v = { 7,8,9 };      // with or without
    tools_log(); cout << "p[1]=" << c.p[1] << ", v[2]=" << v[2] << endl;
    
    D d = { 1.1, 2.2, 4.4 };        // do it in my own class
    tools_log(); cout << "d[0]=" << d.v[0] << ", d[1]=" << d.v[1] << endl;

    return EXIT_SUCCESS;
}

