// Author: A.Voss@FH-Aachen.de
//
// idea
// tags: printf

// content: inheritance, virtual functions
//     new: virtual, override, final, hex, dec, member-initializer

#include <iostream>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

// note: "final" and "override" are C++11-features!

// empty class but with one virtual function
class D
{
public:
    // virtual ~D() {}
    void f1()               { tools_log(); cout << "D::f1" << endl; }
    virtual void f2()       { tools_log(); cout << "D::f2" << endl; }
    virtual void f3(int n)  { tools_log(); cout << "D::f3 n=" << n << endl; }
    virtual void f4() final { tools_log(); cout << "D::f4" << endl; }
};

// empty class inheriting virtual functions;
// class is marked as final thus it is not possible to inherit from
class E final : public D
{
public:
    void f1()               { tools_log(); cout << "E::f1" << endl; }
    void f2()               { tools_log(); cout << "E::f2" << endl; }
    virtual void f3(int n)  { tools_log(); cout << "E::f3 n=" << n << endl; }

// note: overriding needs exact the same parameters ..
// use "override" to emphasize your intention, this results in an error 
// if declarations differ
 //   virtual void f3(unsigned int n)  override  { cout << "E::f3 n=" << n << endl; }

// the following is not possible as f4 is declared as "final"
//    virtual void f4() final { cout << "D::f4" << endl; }
};


int main()
{
    tools_log(); cout << "define objects" << endl;
    D d;
    E e;
    
    tools_log(); cout << "d:" << endl;
    d.f1();
    d.f2();
    tools_log(); cout << "e:" << endl;
    e.f1();
    e.f2();
    
    // this is a D-reference to an E-obj (remember: E is D)
    D & r = e;
    tools_log(); cout << "r:" << endl;
    r.f1();
    r.f2();
    r.f3(1);
   
    // same principle
    tools_log(); cout << "v:" << endl;
    vector<D*> v = { new D(), new E(), new D(), new E() };
    for (D* p : v)
        p->f2();

#pragma GCC diagnostic ignored "-Wdelete-non-virtual-dtor"
    // delete pointers -> this is a problem, why?
    for (D* p : v)
        delete p;
    v.clear(); 
    // consider a virtual descructor for classes being inherited
    // ie. here: add to D: virtual ~D() {}

    return EXIT_SUCCESS;
}

